%%%----------------------------------------------------------------------
%%% File    : mlex.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : Simple lexical scanner
%%% Created :  2 Sep 2002 by  <svg@surnet.ru>
%%%----------------------------------------------------------------------
%%%
%%% $Id$
%%%
%%% $Log$
%%%

-module(mlex).
-author('svg@surnet.ru').

-export([scan/3, scan_token/3, scan_token/2, grammar/1, format_error/1]).
-export([c/1, nc/1, ci/1, ci/2, cni/1, cni/2, str/1, '.'/0]).
-export([match/1, match/2, nmatch/1, nmatch/2]). 
-export(['@'/1, '|'/1, '*'/1, '+'/1, '?'/1, sol/1, eol/1, btw/3]).

-define(ERROR_BEFORE_LEN, 25).
-define(skip, fun (_, _, _) -> [] end).

-record(dfa_pos, {pfun, class, follow, pred}).

				    
%% Scanner

scan(ModBuffer, Buffer, Grammar) ->
  Ret =
    case catch scan1(create_buf(ModBuffer, Buffer), Grammar, []) of
      {ok, Result} ->
	Result;
      Error = {error, _} ->
	Error;
      Other ->
	{error, Other}
    end,
  Ret.

scan1(Buf, Grammar, Tokens) ->
  case scan_token(Buf, Grammar) of
    {eof, Buf1} ->
      {ok, lists:reverse(Tokens)};
    {ok, {Token, Buf1}} ->
      NxtTokens =
	case Token of
	  [] ->
	    Tokens;
	  _ ->
	    [Token|Tokens]
	end,
      scan1(Buf1, Grammar, NxtTokens);
    Error ->
      Error
  end.

scan_token(ModBuffer, Buffer, Grammar) ->
  scan_token(create_buf(ModBuffer, Buffer), Grammar).

scan_token(Buf, Grammar) ->
  case nxtchr(Buf) of
    Ret = {eof, Buf1} ->
      Ret;
    {C, Buf1} ->
      case scan_token(C, Buf1, Grammar, [], [], []) of
	{ok, {[], Buf2}} ->
	  scan_token(Buf2, Grammar);
	Result ->
	  Result
      end
  end.

scan_token(C, Buf, Grammar, Final, Pred, ScanChars) ->
  {Next, Grammar1} = grammar_next(C, Grammar),
  case Next of
    false when C == sol;
	       C == eol ->
      {C1, Buf1} = nxtchr(Buf),
      scan_token(C1, Buf1, Grammar, Final, Pred, ScanChars);
    false ->
      result_token(retchr(C, Buf), Grammar, Final, Pred, ScanChars);
    {_, {_, AddPred}} ->
      {Scan, Final1} =
	case state_final(Next) of
	  false -> {long, Final};
	  FinalFun -> {FinalFun(scan), {buf_pos(Buf), FinalFun}}
	end,
      ScanChars1 =
	case C of
	  _ when C == sof;
		 C == eof;
		 C == sol;
		 C == eol ->
	    ScanChars;
	  _ -> 
	    [C|ScanChars]
	end,

      case Scan of
	short ->
	  result_token(Buf, Grammar1, Final1, Pred, ScanChars1);
	long ->
	  {C1, Buf1} = nxtchr(Buf),
	  Pred1 = add_pred_pos(AddPred, Pred, buf_pos(Buf)),
	  scan_token(C1, Buf1, Grammar1, Final1, Pred1, ScanChars1);
	Other ->
	  {_, F} = Final1,
	  exit({bad_scan_type, F(class)})
      end
  end.

result_token(Buf, Grammar, {StatePos, FinalFun}, Pred, ScanChars) ->
  Pos = get_pred_pos(FinalFun(final), Pred, StatePos),
  Buf1 = buf_pos(Buf, Pos),
  LNum = lnum(Buf1),
  RestBuf = forward(Buf1),
  {Class, Format} = final_params(FinalFun),
  TokenStr = lists:sublist(lists:reverse(ScanChars), Pos),
  {ok, {Format(Class, LNum, TokenStr), RestBuf}};
result_token(Buf, Grammar, [], _, Chars) ->
  token_error(Buf, Grammar, Chars).

grammar_expect({{_, {Funs, _}}, _, _}) ->
  [F(class) || F <- Funs].

grammar_next(C, G={{N, {Funs, _}}, D, DT}) ->
  Success = lists:usort([F(class) || F <- Funs, F(C) == true]),
  case Success of
    [] -> {false, G};
    _ ->
      N1 = dfa_trans_get(N, Success, DT),
      StData = dfa_st_get(N1, D),
      St = {N1, StData},
      {St, {St, D, DT}}
  end.

get_pred_pos(Num, Pred, Default) ->
  case lists:keysearch(Num, 1, Pred) of
    {value, {_, Pos}} ->
	Pos;
      _ ->
      Default
  end.

add_pred_pos(Add, To, Pos) ->
  [{N, Pos} || N <- Add] ++
  [N || N <- To, lists:member(element(1, N), Add) =/= true].

final_params(Fun) ->
  {Fun(class), Fun(funformat)}.

state_final({_, {Funs, _}}) ->
  Funs1 = [F || F <- Funs, F(final) =/= false],
  case Funs1 of
    [] ->
      false;
    _ ->
      first_final(Funs1)
  end.

first_final([F1|Funs]) ->
  lists:foldl(fun (F, PF) ->
		  case (F(final) < PF(final)) of
		    true -> F;
		    _ -> PF
		  end
	      end, F1, Funs).

token_error(Buf, Grammar, Chars) ->
  Chars1 = if length(Chars) > ?ERROR_BEFORE_LEN ->
	       lists:sublist(Chars, ?ERROR_BEFORE_LEN);
	      true ->
	       Chars
	   end,
  Str = lists:reverse(Chars1),
  Expect = grammar_expect(Grammar),
  {Chr, _} = nxtchr(Buf),
  {error, {scan, lnum(Buf), Chr, Expect, Str}}.

format_error({scan, Line, Chr, Expect, Str}) ->
  FmtExp = lists:flatten([format_expected(E) || E <- Expect]),
  lists:flatten(
    io_lib:format("   Scan error at line: ~w;~n"
		  "                 Char: ~c;~n"
		  " Scanned chars before: ~s;~n"
		  "Expected char classes: ~s~n",
		  [Line, Chr, Str, FmtExp]));
format_error(Error) ->
  lists:flatten(io_lib:format("~w", [Error])).


format_expected({not_match, Exp}) ->
  "not " ++ format_expected(Exp);
format_expected(C) when integer(C) ->
  io_lib:format("~c;", [C]);
format_expected({From, To}) when integer(From), integer(To) ->
  io_lib:format("[~c-~c];", [From, To]);
format_expected(Str) when list(Str) ->
  io_lib:format("~s;", [Str]);
format_expected(Exp) ->
  io_lib:format("~w;", [Exp]).

%% Buffer

create_buf(M, B) ->
  {[], {M, B}}.

nxtchr({sol, {M, B}}) ->
  {sol, {[], {M, B}}};
nxtchr({Int, {M, B}}) ->
  {C1, B1} = M:nxtchr(B),
  nxtchr(C1, {Int, {M, B1}}).

nxtchr(C, {[], {M, B}}) when C == $\r; C == $\n ->
  B1 = M:retchr(B),
  {eol, {eol, {M, B1}}};
nxtchr(C, {eol, {M, B}})  when C == $\r ->
  {C1, B1} = M:nxtchr(B),
  B2 = M:retchr(B1),
  case C1 of
    $\n ->
      {C, {eol, {M, B2}}};
    _ ->
      {C, {sol, {M, B2}}}
  end;
nxtchr(C, {eol, {M, B}})  when C == $\n ->
  {C, {sol, {M, B}}};
nxtchr(C, Buf) ->
  {C, Buf}.

retchr(Internal, Buf) when Internal == sof;
			   Internal == eof ->
  Buf;
retchr(Internal, {_, Buf}) when Internal == sol;
				Internal == eol ->
  {Internal, Buf};
retchr(C, {sol, {M, B}}) when C == $\r; C == $\n ->
  B1 = M:retchr(B),
  {eol, {M, B1}};
retchr(eol, {_, {M, B}}) ->
  {[], {M, B}};
retchr(C, {Int, {M, B}}) ->
  B1 = M:retchr(B),
  {Int, {M, B1}}.

forward({Int, {M, B}}) ->
  B1 = M:forward(B),
  {Int, {M, B1}}.

lnum({_, {M, B}}) ->
  M:lnum(B).

buf_pos({_, {M, B}}) ->
  M:pos(B).

buf_pos({[], {M, B}}, Pos) ->
  {[], {M, M:pos(B, Pos)}};
buf_pos(Buf = {Int, {M, B}}, Pos) ->
  case M:pos(B) of
    Pos ->
      Buf;
    P when P == Pos-1 ->
      CurChr = M:curchr(B),
      B1 = M:pos(B, Pos),
      PrvChr = M:curchr(B1),
      case {PrvChr, CurChr} of
	{$\r, $\n} ->
	  {eol, B1};
	_ ->
	  {[], {M, B1}}
      end
  end.
%% Grammar

grammar([]) ->
  grammar(internal_rules());
grammar(Rules) when list(Rules) ->
  compile_grammar('|'(compile_rules(internal_rules() ++ Rules))).

compile_grammar(Grammar) ->
  {First, D, DT} = dfa(Grammar),
  Funs = dfa_st_get(First, D),
  {{First, Funs}, D, DT}.

compile_rules(Ts) ->
  compile_rules(Ts, 0, []).

compile_rules([], _, Ret) ->
  lists:reverse(Ret);
compile_rules([T|Ts], Cnt, Acc) ->
  compile_rules(Ts, Cnt+1, [compile_rule(T, Cnt)|Acc]).

compile_rule({Class, Tree1, '/', Tree2, Format}, Num) ->
  compile_rule({Class, Tree1, '/', Tree2, Format, long}, Num);
compile_rule({Class, Tree1, '/', Tree2, Format, Scan}, Num) ->
  '@'([Tree1, pred(Num), Tree2, final(Class, Format, Num, Scan)]);
compile_rule({Class, Tree, Format}, Num) ->
  compile_rule({Class, Tree, Format, long}, Num);
compile_rule({Class, Tree, Format, Scan}, Num) ->
  '@'([Tree, final(Class, Format, Num, Scan)]).

final(Class, Format, Num, Scan) when Scan == short;
				     Scan == long ->
  Test = fun (final) -> Num;
	     (funformat) -> Format;
	     (class) -> Class;
	     (scan) -> Scan;
	     (_) -> false
	 end,
  fun (D) ->
      tree_pos({Class, Test}, D)
  end.

internal_rules() ->
  [{sof, match(sof), ?skip},
   {sol, match(sol), ?skip},
   {eol, match(eol), ?skip}].

%% Helpers
nmatch(M) ->
  nmatch(M, fun (I) -> I == M end).

nmatch(Class, Test) when function(Test) ->
  match({not_match, Class}, fun (I) -> not Test(I) end).

match(M) ->
  match(M, fun (I) -> I == M end).
match(Class, TestFun) when function(TestFun) ->
  Test = fun (class) -> Class;
	     (final) -> false;
	     (eof) -> false;
	     (I) ->
	     TestFun(I)
	 end,
  fun (D) ->
      tree_pos({Class, Test}, D)
  end.

nc(C) when integer(C) ->
  nmatch(C).

c(C) when integer(C) ->
  match(C).

'.'() ->
  cni([sol, eol, $\r, $\n]).

str(Str) when list(Str) ->
  '@'([c(C) || C <- Str]).

ci(Str) when list(Str) ->
  ci(Str, Str).

ci(From, To) when integer(From),
		  integer(To),
		  From =< To ->
  Test =
    fun (C) when integer(C),
		 C >= From,
		 C =< To ->
	true;
	(_) ->
	false
    end,
  match({From, To}, Test);
ci(Class, Str) when list(Str), list(Class) ->
  Test =
    fun (C) -> lists:member(C, Str) end,
  match(Class, Test).

cni(Str) ->
  cni(Str, Str).

cni(From, To) when integer(From),
		   integer(To),
		   From =< To ->
  Test =
    fun (C) when integer(C),
		 C >= From,
		 C =< To ->
	true;
	(_) ->
	false
    end,
  nmatch({From, To}, Test);
cni(Class, Str) when list(Str), list(Class) ->
  Test = fun (C) -> lists:member(C, Str) end,
  nmatch(Class, Test).

sol(Node) ->
  '@'(['|'([match(sof)|match(sol)]), Node]).

eol(Node) ->
  '@'([Node, match(eol)]).

'@'(Nodes) when list(Nodes) ->
  fun (D) ->
      {Ps, D1} = tree_compile(Nodes, D),
      tree_node(cat, Ps, D1)
  end.

'|'(Nodes) when list(Nodes) ->
  fun (D) ->
      {Ps, D1} = tree_compile(Nodes, D),
      tree_node('|', Ps, D1)
  end.

'+'(Node) when function(Node) ->
  '@'([Node, '*'(Node)]).

'?'(Node) when function(Node) ->
  '|'([Node, e()]).

'*'(Node) when function(Node) ->
  fun (D) ->
      {P, D1} = Node(D),
      tree_node('*', P, D1)
  end.

btw(From, To, Node) when integer(From),
			 integer(To),
			 From =< To,
			 From > 0 ->
  btw(From, To, Node, []).

btw(From, To, Node, [SQ]) when To < From ->
  SQ;
btw(From, To, Node, SQ) when To < From ->
  '|'(SQ);
btw(From, To=1, Node, SQ) ->
  btw(From, To-1, Node, [Node|SQ]);
btw(From, To, Node, SQ) ->
  btw(From, To-1, Node, ['@'(lists:duplicate(To, Node))|SQ]).

pred(N) ->
  fun (D) ->
      tree_pos({'$pred', N}, D)
  end.

e() ->
  fun (D) ->
      tree_pos('$e', D)
  end.

tree_compile(Nodes, D) ->
  tree_compile(Nodes, [], D).

tree_compile([], Ps, D) ->
  {lists:reverse(Ps), D};
tree_compile([Node|Rest], Ps, D) ->
  {P, NxtD} = Node(D),
  tree_compile(Rest, [P|Ps], NxtD).

dfa(Dfa) ->
  PsD = {create_counter(), create_dict()},
  {{First, _, _}, {_, PsD1}} = Dfa(PsD),
  DfaFirst = lists:usort(First),
  {D, DT} = dfa1([DfaFirst], PsD1, create_dict(), create_dict()),
  {DfaFirst, D, DT}.

dfa1([], _, D, DT) ->
  {D, DT};
dfa1([[]], _, D, DT) ->
  {D, DT};
dfa1([St|Rest], PsD, D, DT) ->
  {NxtD, NxtDT} = add_dfa_st(St, PsD, D, DT),
  dfa1(Rest, PsD, NxtD, NxtDT).

add_dfa_st(St, PsD, D, DT) ->
  case dfa_st_exist(St, D) of
    true ->
      {D, DT};
    false ->
      Ps = [dfa_pos_get(N, PsD) || N <- St],
      Pred = lists:usort([dfa_pos_pred(P)
			  || P <- Ps, dfa_pos_pred(P) =/= undefined]),
      Funs = [dfa_pos_fun(P) || P <- Ps],
      NxtD = dfa_st_add(St, {Funs, Pred}, D),
      Classes = class_intersects(
		  lists:usort([dfa_pos_class(P) || P <- Ps])),
      Edges = [mk_dfa_edge(Class, Ps) || Class <- Classes],
      NxtDT = dfa_trans_add(St, Edges, DT),
      dfa1([S || {_, S} <- Edges], PsD, NxtD, NxtDT)
  end.

class_intersects(Classes) ->
  class_intersects(Classes, []).

class_intersects([], Res) ->
  [lists:usort(N) || N <- Res];
class_intersects([C|Rest], Is) ->
  Test = fun (C1) -> class_intersect(C, C1) end,
  class_intersects(Rest,
		   [[C]|Is] ++ [[C|I] || I <- Is, lists:any(Test, I) == true]).
						    
class_intersect(C1, C2) ->
  case class_intersect1(C1, C2) of
    false -> class_intersect1(C2, C1);
    _ -> true
  end.

class_intersect1(C, {not_match, C}) ->
  false;
class_intersect1(_, {not_match, _}) ->
  true;
class_intersect1(C, C) ->
  true;
class_intersect1(C, Str) when integer(C), list(Str) ->
  lists:member(C, Str);
class_intersect1(C, {From, To}) when integer(C) ->
  (C >= From) and (C =< To);
class_intersect1(Str1, Str2) when list(Str1), list(Str2) ->
  lists:any(fun (E) ->
		lists:member(E, Str2)
	    end, Str1);
class_intersect1(Str, {From, To}) when list(Str) -> 
  lists:any(fun (E) ->
		class_intersect1(E, {From, To})
	    end, Str);
class_intersect1({From1, To1}, {From2, To2}) ->
  (From2 =< To1) and (To1 =< To2);
class_intersect1(_, _) ->
  false.

%% State table

dfa_st_get(StName, D) ->
  dict_get(D, StName).

dfa_st_add(StName, Data, D) ->
  dict_set(D, StName, Data).

dfa_st_exist(Name, D) ->
  dict_exist(D, Name).

%% Trans table

%%dfa_class_sts(Class, DT) ->
%%  dict_select(DT,
%%	      fun (Key={_, C}, _) when C == Class -> true;
%%		  (_, _) -> false
%%	      end,
%%	      fun ({St, _}, _) ->
%%		  St
%%	      end).

dfa_trans_add(St, [], DT) ->
  DT;
dfa_trans_add(St, [{Class, To}|Edges], DT) ->
  dfa_trans_add(St, Edges, dict_set(DT, {St, Class}, To)).

dfa_trans_get(St, Class, DT) -> 
  dict_get(DT, {St, Class}).

mk_dfa_edge(Class, Ps) ->
  Test = fun (P) -> lists:member(P, Class) end,
  {Class, lists:usort(
	    lists:flatten([dfa_pos_follow(P)
			   || P <- Ps, Test(dfa_pos_class(P)) == true]))}.
%% Tree

tree_pos({'$pred', N}, D) ->
  {{[], [], {pred, N}}, D};
tree_pos('$e', D) ->
  {{[], [], true}, D};
tree_pos({Class, Test}, D) ->
  {Pos, D1} = new_pos(Class, Test, D),
  Last = First = [Pos],
  Null = false,
  {{First, Last, Null}, D1}.

tree_node('|', [P1, P2|Ps], D) ->
  tree_node('|', [P2|Ps], P1, D);
tree_node(cat, [P1, P2|Ps], D) ->
  tree_node(cat, [P2|Ps], P1, D);
tree_node('*', {First, Last, _}, D) ->
  NxtD = add_follow(Last, First, D),
  {{First, Last, true}, NxtD}.

tree_node('|', [], Res, D) ->
  {Res, D};
tree_node('|', [Pos|Rest], {First, Last, Null}, D) ->
  tree_node('|', Rest,
	   {firstpos(Pos) ++ First, lastpos(Pos) ++ Last, (nullable(Pos) or Null)},
	   D);
tree_node(cat, [], Res, D) ->
  {Res, D};
%% Predictive node
tree_node(cat, [{[], [], {pred, N}}, Pos|Rest], Prev, D) ->
  NxtD = add_pred(firstpos(Pos), N, D),
  tree_node(cat, [Pos|Rest], Prev, NxtD);
tree_node(cat, [Pos|Rest], {First, Last, Null}, D) ->
  NxtFirst = case Null of
	       true ->
		 firstpos(Pos) ++ First;
	       false ->
		 First
	     end,
  NxtLast = case nullable(Pos) of
	      true ->
		lastpos(Pos) ++ Last;
	      false ->
		lastpos(Pos)
	     end,
  NxtNull = (nullable(Pos) and Null),
  NxtD = add_follow(Last, firstpos(Pos), D),
  tree_node(cat, Rest, {NxtFirst, NxtLast, NxtNull}, NxtD).

new_pos(Class, Test, D) ->
  {Name, NxtD} = add_pos(mk_dfa_pos(Test, Class, []), D).

firstpos({First, _, _}) ->  
  First.
lastpos({_, Last, _}) ->
  Last.

nullable({_, _, Null}) ->
  Null.

add_follow([], _, D) ->
  D;
add_follow([Name|Rest], Follow, {C, D}) ->
  D1 = dfa_pos_follow(Name, Follow, D),
  add_follow(Rest, Follow, {C, D1}).

add_pred([], _, D) ->
  D;
add_pred([Name|Rest], N, {C, D}) ->
  D1 = dfa_pos_pred(Name, N, D),
  add_pred(Rest, N, {C, D1}).

add_pos(Data, {C, D}) ->
  {Name, NxtC} = counter_next(C),
  NxtD = dict_set(D, Name, Data),
  {Name, {NxtC, NxtD}}.

%% Pos

mk_dfa_pos(Test, Class, Follow) ->
  #dfa_pos{pfun=Test, class=Class, follow=Follow}.

dfa_pos_get(Name, D) ->
  dict_get(D, Name).

dfa_pos_set(Name, Data, D) ->
  dict_set(D, Name, Data).

dfa_pos_class(P=#dfa_pos{class=Class}) ->
  Class.

dfa_pos_fun(P=#dfa_pos{pfun=Fun}) ->
  Fun.

dfa_pos_pred(Name, N, D) ->
  P = dfa_pos_get(Name, D),
  dfa_pos_set(Name, P#dfa_pos{pred=N}, D).

dfa_pos_pred(P=#dfa_pos{pred=N}) ->
  N.

dfa_pos_follow(Name, Follow, D) ->
  P = dfa_pos_get(Name, D),
  dfa_pos_set(Name,
	      P#dfa_pos{follow=lists:usort(dfa_pos_follow(P) ++ Follow)}, D).

dfa_pos_follow(P=#dfa_pos{follow=Follow}) ->
  Follow.

%% Counter
create_counter() ->
  0.

counter_next(C) ->
  {C+1, C+1}.

%% Dictionary

dict_get(Dict, Name) ->
  gb_trees:get(Name, Dict).

dict_set(Dict, Name, Val) ->
  NewDict = gb_trees:enter(Name, Val, Dict).

dict_exist(Dict, Name) ->
  gb_trees:is_defined(Name, Dict).

%%dict_to_list(Dict) ->
%%  gb_trees:to_list(Dict).

%%dict_select(Dict, Test, Val) ->
%%  dict_select(Test, Val, [], gb_trees:next(gb_trees:iterator(Dict))).
%%
%%dict_select(Test, Val, Res, none) ->
%%  lists:reverse(Res);
%%dict_select(Test, Val, Acc, {K, V, NxtS}) ->
%%  case Test(K, V) of
%%    true ->
%%      dict_select(Test, Val, [Val(K, V)|Acc], gb_trees:next(NxtS));
%%    false ->
%%      dict_select(Test, Val, Acc, gb_trees:next(NxtS))
%%  end.

create_dict() ->
  gb_trees:empty().
