%%%----------------------------------------------------------------------
%%% File    : mlex_test.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : 
%%% Created :  2 Sep 2002 by  <svg@surnet.ru>
%%%----------------------------------------------------------------------
%%%
%%% $Id$
%%%
%%% $Log$
%%%

-module(mlex_test).
-author('svg@surnet.ru').

-export([test_all/0, test_all/1]).
-export([test_file_scan/1, test_predictive/1, test_short/1, test_scan_cc/1]).

-import(mlex,[c/1, nc/1, ci/1, ci/2, cni/1, cni/2, str/1, btw/3]).
-import(mlex,['@'/1, '|'/1, '*'/1, '+'/1, '?'/1, '.'/0, sol/1, eol/1]).

-define(match(_Pat), fun (_Res) ->
                         case _Res of
                           _Pat -> ok;
                           _ -> false
                         end
                     end).

test_all([quit]) ->
  (catch test_all()),
  halt().

test_all() ->
  lists:foreach(fun(T) ->
                    io:format("~n~s~n",[apply({?MODULE, T}, [info])]),
                    apply({?MODULE, T}, [do])
                end,
                [test_predictive, test_file_scan, test_scan_cc, test_short]
               ).

test_file_scan(info) ->
  "*** Scan file ***";
test_file_scan(do) ->
  FileAsn1 = "./priv/test/CISCO-AAA-SESSION-MIB.txt",
  FileConf = "./priv/test/freeradius.txt",
  Tests =
    [{"ASN.1", [FileAsn1, mlex_asn1:grammar()],
      ?match([{upper,12,"CISCO-AAA-SESSION-MIB"}|_])},
     {"FreeRadius config", [FileConf, mlex_freeradius_conf:grammar()],
      ?match([{decl_end,2,"\n"}|_])}
    ],
  process_tests(fun scan_file/2, Tests).

test_scan_cc(info) ->
  "*** Scan file with continuation ***";
test_scan_cc(do) ->
  BufConf = file_to_buf("./priv/test/freeradius.txt"),
  Tests =
    [{"FreeRadius config", [BufConf, mlex_freeradius_conf:grammar()],
      ?match([{decl_end,2,"\n"}|_])}
    ],
  process_tests(fun scan_cc/2, Tests).

test_short(info) ->
  "*** Short expression ***";
test_short(do) ->
  Short = mlex:grammar([{short, '+'(str("abba")), fun yeec_token/3, short},
			{long, '+'(str("abba")), fun yeec_token/3}]),
  Long = mlex:grammar([{short, '+'(str("abba")), fun yeec_token/3, short},
		       {long, '+'(str("abba")), fun yeec_token/3}]),
  Str = "abbaabbaabbaabba",
  Tests =
    [{"short first", [Str, Short],
      ?match([{short,1,"abba"},
	      {short,1,"abba"},
	      {short,1,"abba"},
	      {short,1,"abba"}])},
     {"long first", [Str, Long],
      ?match([{short,1,"abba"},
	      {short,1,"abba"},
	      {short,1,"abba"},
	      {short,1,"abba"}])}
    ],
  process_tests(fun scan_string/2, Tests).

test_predictive(info) ->
  "*** Predictive operator ***";
test_predictive(do) ->
  StrTrue = "abcdeabcdede",
  StrFalse = "abcdeabcdedeabcabc",
  Tests =
    [{"must be true", [StrTrue, predictive_grammar()],
      ?match([{abc,1,"abc"}, {de,1,"de"}, {abc,1,"abc"},
	      {de,1,"de"}, {de,1,"de"}])},
     {"must be false", [StrFalse, predictive_grammar()],
      ?match("   Scan error at line: 1;\n"
	     "                 Char: a;\n"
	     " Scanned chars before: abc;\n"
	     "Expected char classes: d;\n")}
    ],
  process_tests(fun scan_string/2, Tests).

predictive_grammar() ->
  mlex:grammar([{abc, str("abc"), '/', '+'(str("de")), fun yeec_token/3},
		{de, str("de"), fun yeec_token/3}]).

scan(Buf, Grammar) ->
  case mlex:scan(mlex_str_buf, Buf, Grammar) of
    {error, Error} ->
      mlex:format_error(Error);
    Res ->
      Res
  end.

scan_cc(Buf, Grammar) ->
  scan_cc(first_token(Buf, Grammar), Grammar, []).

scan_cc({error, Error}, Grammar, Tokens) ->
  mlex:format_error(Error);
scan_cc({eof, _}, _, Res) ->
  lists:reverse(Res);
scan_cc({ok, {Token, Buf}}, Grammar, Tokens) ->
  scan_cc(next_token(Buf, Grammar), Grammar, [Token|Tokens]).

first_token(Buf, Grammar) ->
  mlex:scan_token(mlex_str_buf, Buf, Grammar).

next_token(Buf, Grammar) ->
  mlex:scan_token(Buf, Grammar).

scan_file(FileName, Grammar) ->
  Buf = file_to_buf(FileName),
  scan(file_to_buf(FileName), Grammar).

scan_string(Str, Grammar) ->
  scan(string_to_buf(Str), Grammar).

file_to_buf(FileName) ->
  {ok, Bin} = file:read_file(FileName),
  string_to_buf(Bin).

string_to_buf(Bin) when binary(Bin) ->
  string_to_buf(binary_to_list(Bin));
string_to_buf(Str) ->
  mlex_str_buf:str_to_buf(Str).

yeec_token(Class, Line, Data) ->
  {Class, Line, Data}.

%% Test help funs

process_tests(Tests) ->
  lists:foreach(fun ({Fun, FunTests}) ->
                    process_tests(Fun, FunTests)
                end, Tests).

process_tests(Fun, Tests) ->
  process_tests(Fun, Tests, 1).

%% Test argument is in the form {Msg, Args, TrueResult}
%% Msg        - test purpose
%% Args       - [Arg]
%% TrueResult - true result pattern
process_tests(Fun, [], _) ->
  ok;
process_tests(Fun, [Test|Tail], Num) ->
  {Msg, Args, ResPat} = Test,
  Status = case catch apply(Fun,Args) of
             Res when function(ResPat) ->
               case catch ResPat(Res) of
                 ok ->
                   [passed, ok];
                 _ ->
                   [Res, failed]
               end;
             ResPat ->
               [passed, ok];
             Other ->
               [Other, failed]
           end,
  io:format("Test ~2.2w: "++Msg++" -> ~p, ~p~n", [Num|Status]),
  process_tests(Fun, Tail, Num+1).
